#include <stdio.h>
#include <stdlib.h>
#include <gccore.h>
#include <unistd.h>
#include <ogc/machine/processor.h>
#include <ogc/stm.h>
#include <ogc/system.h>
#include <wiiuse/wpad.h>
#include <sdcard/wiisd_io.h>
#include <fat.h>

static void *xfb = NULL;
static GXRModeObj *rmode = NULL;

static vu16* const viReg = (u16*)0xCC002000;
int shutdown = 0;

void powerbtn() {
	shutdown = 1;
}

void printVi(u16* v) {
	printf(" VTiming:   %04x                  Control:   %04x              \n",
	                    v[ 0],                       v[ 1]                  );
	printf(" HTiming:   %04x %04x  %04x %04x  VBlank:    %04x %04x  %04x %04x\n",
	                    v[ 2],v[ 3],v[ 4],v[ 5],       v[ 6],v[ 7],v[ 8],v[ 9]);
	printf(" Burst:     %04x %04x  %04x %04x                               \n",
	                    v[10],v[11],v[12],v[13]                               );
	printf(" XfbInfo:   %04x %04x  %04x %04x             %04x %04x  %04x %04x\n",
	                    v[14],v[15],v[16],v[17],v[18],v[19],v[20],v[21]       );
	printf(" VBeam:     %04x                  HBeam:     %04x\n",
	                    v[22],         v[23]          );
	printf(" PreTrace:  %04x %04x             PostTrace: %04x %04x\n",
	                    v[24],v[25],          v[26],v[27] );
	printf(" Interrupt: %04x %04x  %04x %04x  Latch:     %04x %04x  %04x %04x\n",
	                    v[28],v[29],v[30],v[31], v[32],v[33],v[34],v[35]        );
	printf(" PictConf:  %04x                  HScaleR:   %04x           \n",
	                 v[36],v[37]                                            );
	printf(" Filter:    %04x %04x  %04x %04x  %04x %04x  %04x %04x      \n",
	                    v[38],v[39],v[40],v[41],v[42],v[43],v[44],v[45]        );
	printf(" Filter:    %04x %04x  %04x %04x  %04x %04x                 \n",
	                    v[46],v[47],v[48],v[49],v[50],v[51]                    );
	printf(" UnkAA:     %04x %04x      Clock: %04x  DTV: %04x           \n",
	                 v[52],v[53],      v[54],     v[55]                     );
	printf(" FBWidth:   %04x     BorderBlank: %04x %04x   Nothing?: %04x\n",
	                  v[56],            v[57],v[58],       v[59]            );
}

void copyVi(u16* v) {
	u32 level;
	_CPU_ISR_Disable(level);
	
	for (u32 i = 0; i < 60; ++i) {
		v[i] = viReg[i];
	}
	
	_CPU_ISR_Restore(level);
}

int main(int argc, char **argv) {
	u16 viAtBoot[60];
	//u16 viNow[60];
	copyVi(viAtBoot);

	// Initialise the video system
	VIDEO_Init();

	// This function initialises the attached controllers
	WPAD_Init();
	
	WPAD_SetIdleTimeout(120);
	WPAD_SetPowerButtonCallback(&powerbtn);
	SYS_SetPowerCallback(&powerbtn);

	// Obtain the preferred video mode from the system
	// This will correspond to the settings in the Wii menu
	rmode = VIDEO_GetPreferredMode(NULL);

	// Allocate memory for the display in the uncached region
	xfb = MEM_K0_TO_K1(SYS_AllocateFramebuffer(rmode));

	// Initialise the console, required for printf
	console_init(xfb,20,20,rmode->fbWidth,rmode->xfbHeight,rmode->fbWidth*VI_DISPLAY_PIX_SZ);

	// Set up the video registers with the chosen mode
	VIDEO_Configure(rmode);

	// Tell the video hardware where our display memory is
	VIDEO_SetNextFramebuffer(xfb);

	// Make the display visible
	VIDEO_SetBlack(FALSE);

	// Flush the video register changes to the hardware
	VIDEO_Flush();

	// Wait for Video setup to complete
	VIDEO_WaitVSync();
	if(rmode->viTVMode&VI_NON_INTERLACE) VIDEO_WaitVSync();

	u64 time = SYS_Time();
	s32 region = CONF_GetRegion();
	s32 videoMode = CONF_GetVideo();
	s32 progScanEnabled = CONF_GetProgressiveScan();
	s32 pal60enabled = CONF_GetEuRGB60();
	s32 aspectRatio = CONF_GetAspectRatio();
	u32 dtv = viAtBoot[55];

	s8 tmp;
	s32 displayOffset = CONF_GetDisplayOffsetH(&tmp);
	if (displayOffset == 0) {
		displayOffset = (s32)tmp;
	}
	
	char filename[512];
	sprintf(filename, "sd:/vi_region%ld_video%ld_dtv%lu_progscan%ld_60hzpal%ld_aspect%ld_hoffset%ld_%llu.bin",
	                  region, videoMode, dtv, progScanEnabled, pal60enabled, aspectRatio, displayOffset, time);
    //sprintf(filename, "sd:/a.bin");
	int filewritesuccess = 0;
	
	fatUnmount("sd:/");
	__io_wiisd.shutdown();
	if (fatMountSimple("sd", &__io_wiisd)) {
		FILE* fp = fopen(filename, "wb");
		if (fp) {
			if (fwrite(&viAtBoot, 120, 1, fp) == 1) {
				filewritesuccess = 1;
			}
			fclose(fp);
			fp = NULL;
		} else {
			filewritesuccess = 2;
		}
		fatUnmount("sd:/");
		__io_wiisd.shutdown();
	} else {
		filewritesuccess = 3;
	}
	
	u64 framecount = 0;
	while (shutdown == 0) {
		VIDEO_ClearFrameBuffer(rmode, xfb, 0x800080);

		WPAD_ScanPads();
		u32 pressed = WPAD_ButtonsDown(0);

		//copyVi(viNow);
		printf("\x1b[8;0H");
		printVi(viAtBoot);
		//printf("\n");
		//printVi(viNow);
		
		printf("\n\n  %s %s (%d)", filewritesuccess == 1 ? "Written to" : "Failed writing to", filename, filewritesuccess);
		printf("\n\n  frame: %llu", framecount); // dolphin doesn't render new frame if the frame data doesn't change...

		VIDEO_SetNextFramebuffer(xfb);
		VIDEO_Flush();
		VIDEO_WaitVSync();
		++framecount;

		if ( pressed > 0 ) {
			shutdown = -1;
		}
	}

	if (shutdown == 1) {
		SYS_ResetSystem(SYS_POWEROFF, 0, 0);
	} else {
		void (*rld)() = (void(*)()) 0x80001800;
		rld();
	}
	
	while (1) {
		usleep(1000);
	}

	return 0;
}
