﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;

namespace ConsoleApp4
{
	class Program
	{
		static void PrintReg(SortedSet<ushort> values, int memloc, string name)
		{
			Console.Write("u16 at " + memloc.ToString("X2") + " (" + name + "): " + values.Count + " values");
			if (values.Count < 10)
			{
				Console.Write(": ");
				foreach (ushort u in values)
				{
					Console.Write(u.ToString("X4") + "; ");
				}
			}
			Console.WriteLine();
		}

		static void GroupBy(int location, int bytecount, string groupname, List<(string path, byte[] data)> data)
		{
			string dir = @"c:\wiivideomode\vi-dumps-grouped\";
			foreach (var d in data)
			{
				List<byte> b = new List<byte>();
				for (int i = 0; i < bytecount; ++i)
				{
					b.Add(d.data[location + i]);
				}
				string group = BitConverter.ToString(b.ToArray()).Replace("-", "");
				string finaldir = Path.Combine(dir, location.ToString("X2") + "_" + bytecount.ToString("D2") + "bytes_" + groupname, group);
				Directory.CreateDirectory(finaldir);
				File.WriteAllBytes(Path.Combine(finaldir, d.path), d.data);
			}
		}

		static void PrintHtiming1(int location, int bytecount, string groupname, List<(string path, byte[] data)> data)
		{
			SortedDictionary<uint, string> vals = new SortedDictionary<uint, string>();
			foreach (var d in data)
			{
				uint val = BitConverter.ToUInt32(d.data, location).SwapEndian();
				if (!vals.ContainsKey(val))
				{
					vals.Add(val, d.path);
				}
			}

			PrintHtiming1internal(vals, false, true);
			PrintHtiming1internal(vals, false, false);
			PrintHtiming1internal(vals, true, true);
			PrintHtiming1internal(vals, true, false);

			return;
		}

		private static void PrintHtiming1internal(SortedDictionary<uint, string> vals, bool printpal50, bool printar43)
		{
			foreach (var v in vals)
			{
				// vi_region2_video1_dtv1_progscan0_60hzpal0_aspect0_hoffset0_35937235372500000
				bool pal50 = v.Value.Substring(v.Value.IndexOf("_60hzpal") + "_60hzpal".Length, 1) == "0";
				bool ar43 = v.Value.Substring(v.Value.IndexOf("_aspect") + "_aspect".Length, 1) == "0";
				int hoffset = int.Parse(v.Value.Split(new char[] { '_' }).Where(x => x.StartsWith("hoffset")).First().Substring("hoffset".Length));

				if (pal50 == printpal50 && ar43 == printar43)
				{
					Console.WriteLine(v.Key.ToString("X8") + ": Offset " + (hoffset / 2) + ", " + (pal50 ? "50" : "60") + "Hz, " + (ar43 ? "4:3" : "16:9"));
				}
			}
		}

		static void Main(string[] args)
		{
			string path = @"c:\wiivideomode\vi-dumps\";

			List<(string path, byte[] data)> files = new List<(string path, byte[] data)>();

			List<SortedSet<ushort>> values = new List<SortedSet<ushort>>();
			for (int i = 0; i < 60; ++i)
			{
				values.Add(new SortedSet<ushort>());
			}

			foreach (string p in Directory.EnumerateFiles(path, "*", SearchOption.AllDirectories))
			{
				files.Add((Path.GetFileName(p), File.ReadAllBytes(p)));
				using (var fs = new FileStream(p, FileMode.Open))
				{
					for (int i = 0; i < 60; ++i)
					{
						values[i].Add(fs.ReadUInt16().SwapEndian());
					}
				}
			}

			PrintReg(values[0], 0 * 2, "VTiming");
			PrintReg(values[1], 1 * 2, "Control");
			PrintReg(values[2], 2 * 2, "HTiming");
			PrintReg(values[3], 3 * 2, "HTiming");
			PrintReg(values[4], 4 * 2, "HTiming");
			PrintReg(values[5], 5 * 2, "HTiming");
			PrintReg(values[6], 6 * 2, "VBlank");
			PrintReg(values[7], 7 * 2, "VBlank");
			PrintReg(values[8], 8 * 2, "VBlank");
			PrintReg(values[9], 9 * 2, "VBlank");
			PrintReg(values[10], 10 * 2, "Burst");
			PrintReg(values[11], 11 * 2, "Burst");
			PrintReg(values[12], 12 * 2, "Burst");
			PrintReg(values[13], 13 * 2, "Burst");
			PrintReg(values[14], 14 * 2, "XfbInfo");
			PrintReg(values[15], 15 * 2, "XfbInfo");
			PrintReg(values[16], 16 * 2, "XfbInfo");
			PrintReg(values[17], 17 * 2, "XfbInfo");
			PrintReg(values[18], 18 * 2, "XfbInfo");
			PrintReg(values[19], 19 * 2, "XfbInfo");
			PrintReg(values[20], 20 * 2, "XfbInfo");
			PrintReg(values[21], 21 * 2, "XfbInfo");
			PrintReg(values[22], 22 * 2, "VBeam");
			PrintReg(values[23], 23 * 2, "HBeam");
			PrintReg(values[24], 24 * 2, "PreTrace");
			PrintReg(values[25], 25 * 2, "PreTrace");
			PrintReg(values[26], 26 * 2, "PostTrace");
			PrintReg(values[27], 27 * 2, "PostTrace");
			PrintReg(values[28], 28 * 2, "Interrupt");
			PrintReg(values[29], 29 * 2, "Interrupt");
			PrintReg(values[30], 30 * 2, "Interrupt");
			PrintReg(values[31], 31 * 2, "Interrupt");
			PrintReg(values[32], 32 * 2, "Latch");
			PrintReg(values[33], 33 * 2, "Latch");
			PrintReg(values[34], 34 * 2, "Latch");
			PrintReg(values[35], 35 * 2, "Latch");
			PrintReg(values[36], 36 * 2, "PictConf");
			PrintReg(values[37], 37 * 2, "HScaleR");
			PrintReg(values[38], 38 * 2, "Filter0");
			PrintReg(values[39], 39 * 2, "Filter0");
			PrintReg(values[40], 40 * 2, "Filter1");
			PrintReg(values[41], 41 * 2, "Filter1");
			PrintReg(values[42], 42 * 2, "Filter2");
			PrintReg(values[43], 43 * 2, "Filter2");
			PrintReg(values[44], 44 * 2, "Filter3");
			PrintReg(values[45], 45 * 2, "Filter3");
			PrintReg(values[46], 46 * 2, "Filter4");
			PrintReg(values[47], 47 * 2, "Filter4");
			PrintReg(values[48], 48 * 2, "Filter5");
			PrintReg(values[49], 49 * 2, "Filter5");
			PrintReg(values[50], 50 * 2, "Filter6");
			PrintReg(values[51], 51 * 2, "Filter6");
			PrintReg(values[52], 52 * 2, "UnkAA");
			PrintReg(values[53], 53 * 2, "UnkAA");
			PrintReg(values[54], 54 * 2, "Clock");
			PrintReg(values[55], 55 * 2, "DTV");
			PrintReg(values[56], 56 * 2, "FBWidth");
			PrintReg(values[57], 57 * 2, "BorderBlank");
			PrintReg(values[58], 58 * 2, "BorderBlank");
			PrintReg(values[59], 59 * 2, "Nothing?");

			PrintHtiming1(0x08, 4, "HTiming1", files);

			for (int i = 0; i < 60; ++i)
			{
				//GroupBy(i * 2, 2, "u16 at " + (i * 2).ToString("X2"), files);
			}
			GroupBy(0x00, 2, "VTiming", files);
			GroupBy(0x02, 2, "Control", files);
			GroupBy(0x04, 4, "HTiming0", files);
			GroupBy(0x08, 4, "HTiming1", files);
			GroupBy(0x0C, 8, "VBlank", files);
			GroupBy(0x14, 8, "Burst", files);
			GroupBy(0x1C, 16, "XfbInfo", files);
			GroupBy(0x2C, 2, "VBeam", files);
			GroupBy(0x2E, 2, "HBeam", files);
			GroupBy(0x30, 4, "PreTrace", files);
			GroupBy(0x34, 4, "PostTrace", files);
			GroupBy(0x38, 8, "Interrupt", files);
			GroupBy(0x40, 8, "Latch", files);
			GroupBy(0x48, 2, "PictConf", files);
			GroupBy(0x4A, 2, "HScaleR", files);
			GroupBy(0x4C, 28, "Filter", files);
			GroupBy(0x68, 4, "UnkAA", files);
			GroupBy(0x6C, 2, "Clock", files);
			GroupBy(0x6E, 2, "DTV", files);
			GroupBy(0x70, 2, "FBWidth", files);
			GroupBy(0x72, 4, "BorderBlank", files);
			GroupBy(0x76, 2, "Nothing", files);

			return;
		}

	}

	static class Util
	{
		public static Int16 SwapEndian(this Int16 x)
		{
			return (Int16)SwapEndian((UInt16)x);
		}
		public static UInt16 SwapEndian(this UInt16 x)
		{
			return x = (UInt16)
					   ((x << 8) |
						(x >> 8));
		}

		public static Int32 SwapEndian(this Int32 x)
		{
			return (Int32)SwapEndian((UInt32)x);
		}
		public static UInt32 SwapEndian(this UInt32 x)
		{
			return x = (x << 24) |
					  ((x << 8) & 0x00FF0000) |
					  ((x >> 8) & 0x0000FF00) |
					   (x >> 24);
		}
		public static ushort ReadUInt16(this Stream s)
		{
			int b1 = s.ReadByte();
			int b2 = s.ReadByte();

			return (ushort)(b2 << 8 | b1);
		}
	}
}
